; EHCI capabilities registers

EHCI_HCS_NPORTS_MASK = 0xF
EHCI_HCS_PORT_POWER_CONTROL = 0x10

struc EHCI_CAP
{  .caplength       db ?
   .reserved        db ?
   .hciversion      dw ?
   .hcsparams       dd ?
   .hccparams       dd ?
   .hcsp_portroute  dq ?
}
virtual at 0
  ECHI_CAP EHCI_CAP
end virtual

; EHCI operational registers

EHCI_OP_RS = 1
EHCI_OP_PERIODIC_SCHED_EN = 1 shl 4
EHCI_OP_ASYNC_SCHED_EN = 1 shl 5
EHCI_OP_PERIODIC_SCHED_STAT = 1 shl 14
EHCI_OP_ASYNC_SCHED_STAT = 1 shl 15

struc EHCI_OP
{  .usbcmd                dd ?
   .usbsts                dd ?
   .usbintr               dd ?
   .frindex               dd ?
   .ctrldssegment         dd ?
   .periodiclistbase      dd ?
   .asynclistaddr         dd ?
   .reserved              dd dup(0x3f-0x1c)
   .configflag            dd ?
; start of PORTSC
   .portsc	              dd ?
}
virtual at 0
  EHCI_OP EHCI_OP
end virtual

; EHCI QT descriptor

EHCI_QTD_TERMINATE = 1
EHCI_QTD_HALTED = 1 shl 6
EHCI_QTD_ACTIVE = 1 shl 7
; ----------------------
EHCI_QTD_PID_SHIFT = 8
EHCI_QTD_PID = 3 shl EHCI_QTD_PID_SHIFT
; ----------------------
EHCI_QTD_CERR_SHIFT = 10
EHCI_QTD_CERR = 3 shl EHCI_QTD_CERR_SHIFT
; ----------------------
EHCI_QTD_CPAGE_SHIFT = 12
EHCI_QTD_CPAGE = 7 shl EHCI_QTD_CPAGE_SHIFT
; ----------------------
EHCI_QTD_TOTAL_LEN_SHIFT = 16
EHCI_QTD_TOTAL_LEN = (1 shl 15 - 1) shl EHCI_QTD_TOTAL_LEN_SHIFT
; ----------------------
EHCI_QTD_TOGGLE_SHIFT = 31
EHCI_QTD_TOGGLE = 1 shl EHCI_QTD_TOGGLE_SHIFT
; ----------------------
EHCI_QTD_TOGGLE_DATA0 = 0
EHCI_QTD_TOGGLE_DATA1 = 1 shl EHCI_QTD_TOGGLE_SHIFT

struc EHCI_QTD
{  .next_qtd	          dd ?
   .alt_next_qtd          dd ?
   .token                dd ?
   .bufptrs              dd dup(5)
   .bufptrs64            dd dup(5)
}

virtual at 0
  EHCI_QTD EHCI_QTD
end virtual

; EHCI Queue Head

struc EHCI_QH
{  .horiz_link          dd ?
   .epchar              dd ?
   .epcaps              dd ?
   .current_td          dd ?
   ; EHCI_QTD overlay
}

virtual at 0
  EHCI_QH EHCI_QH
end virtual

struc EHCI
{  .cap        dd ? ; capabilities registers address
   .op         dd ? ; operational registers address
   .dummy_qh   dd ? ; "dummy qh" address
   .base       dd ? ; MMIO base address
   .bus        dd ? ; PCI bus
   .devfn      dd ? ; PCI device and function number
}

virtual at 0
  EHCI EHCI
end virtual

; Start EHCI controller
; ---------------------
; ehci_start(hci_t *controller)
proc ehci_start stdcall, hci:dword
       mov      esi, [hci]
	   mov      esi, [esi+EHCI.op]
	   mov      edi, [esi+EHCI_OP.usbcmd]
	   or       edi, EHCI_HC_OP_RS
	   mov      [esi+EHCI_OP.usbcmd], edi
	   ret
endp

; Stop EHCI controller
; --------------------
; ehci_stop(hci_t *controller)
proc ehci_stop stdcall, hci:dword
       mov      esi, [hci]
	   mov      esi, [esi+EHCI.op]
	   mov      edi, [esi+EHCI_OP.usbcmd]
	   and      edi, not EHCI_HC_OP_RS
	   mov      [esi+EHCI_OP.usbcmd], edi
	   ret
endp

; Reset EHCI controller
; ---------------------
; ehci_reset(hci_t *controller)
proc ehci_reset stdcall, hci:dword
		stdcall ehci_stop, [hci]
		; delay 10ms
        mov     esi, [hci]
		mov     esi, [esi+EHCI.op]
		mov     edi, [esi+EHCI_OP.usbsts]
		test    edi, (1 shl 12) ; test of HC_HALTED bit
		jnz     .end
		mov     edi, 2 ; HC_RESET bit
		mov     [esi+EHCI_OP.usbcmd], edi
		; now wait 100ms
		mov     ecx, 10

	.loop:
        ; delay 10ms
		mov     edi, [esi+EHCI_OP.usbcmd]
		test    edi, 2
		jz      .end
		loop    .loop

	.end:
		ret
endp

; Shutdown EHCI controller
; ------------------------
; ehci_shutdown(hci_t *controller)
proc ehci_shutdown stdcall, hci:dword
		mov esi, [hci]
		mov esi, [esi+HCI.data] ; EHCI instance
		stdcall ehci_set_periodic_schedule, esi, 0

		; free periodic frame list
		mov edi, [esi+EHCI.op]
		stdcall free, [edi+EHCI_OP.periodiclistbase]

		; free dummy QH
		stdcall free, [esi+EHCI.dummy_qh]]
		mov [edi+EHCI_OP.configflag], 0
		ret
endp

; Take ownership of EHCI controller
; ---------------------------------
; ehci_take(hci_t *controller)
proc ehci_take stdcall, hci:dword
endp

; Create EHCI structure
; ---------------------
; hci_t *ehci_create()
proc ehci_create stdcall
		; allocate memory for new controller HCI
		mov     esi, [hci]
		mov     esi, [esi+HCI.data]
		; allocate memory for new EHCI instance
		; and write pointer at HCI.data

; Set MMIO access

		stdcall PciRead32, [esi+EHCI.bus], [esi+EHCI.devfn], PCI_COMMAND
		or      eax, PCI_COMMAND_MEMORY
		and     eax, not PCI_COMMAND_IO
		stdcall PciWrite32, [esi+EHCI.bus], [esi+EHCI.devfn], PCI_COMMAND, eax
; Fill controller structure with EHCI function pointers
		;...
; Read MMIO base
		stdcall PciRead32, [esi+EHCI.bus], [esi+EHCI.devfn], USBBASE
		mov     [esi+EHCI.base], eax

		; Initialize 0 device
		stdcall usb_init_device, [ehci], 0
        mov     eax, [esi+EHCI.base]
		mov     [esi+EHCI.cap], eax
		mov     bl, [esi+EHCI.cap] ; read caplength
		add     eax, bl
		mov     [esi+EHCI.op], eax
		ret
endp

FLADJ = 0x61

; Do EHCI initialization
; ----------------------
; ehci_init(hci_t *controller)
proc ehci_init stdcall, hci:dword
		mov     esi, [hci]
		mov     esi, [esi+HCI.data] ; This is EHCI controller
		stdcall PciWrite8, [esi+EHCI.bus], [esi+EHCI.devfn], FLADJ, (60000 - 59488)/6
		stdcall ehci_start, [esi+HCI.data]
		mov     eax, 1
		mov     edi, [esi+EHCI.op]
		mov     [edi+EHCI_OP.configflag], eax ; take control over ports
; Enable USB interrupts (not implemented)
; Define periodic list (not yet implemented)
; Fill dummy qh
		; fill it by zeroes
		mov     edi, [esi+EHCI.dummy_qh]
		mov     [edi+EHCI_QH.horiz_link], EHCI_QH_TERMINATE
		mov     [edi+EHCI_QH.td + EHCI_QTD.next_qtd], EHCI_QTD_TERMINATE
		mov     [edi+EHCI_QH.td + EHCI_QTD.alt_next_qtd], EHCI_QTD_TERMINATE
; Fill periodic list by dummy QHs
; Enable periodic shedule
; Init root hub
        ret
endp

; Find closest USB2.0 hub
; -----------------------
; ehci_closest_usb2_hub(subdev_t *dev, int addr, int port)
proc ehci_closest_usb2_hub stdcall, dev:dword, addr:dword, port:dword
endp

; ---------------------------------------------------------------
;
;                        Root Hub stuff

struc EHCI_RH
{  .ports_num   dd ? ; number of ports
   .ports       dd ? ; pointer to ports address
   .devices     dd ? ; pointer to attached devices
}

virtual at 0
   EHCI_RH EHCI_RH
end virtual

; Initialization of Root Hub
; --------------------------
; ehci_rh_init(usbdev_t *dev)
proc ehci_rh_init stdcall, dev:dword
		mov esi, [dev]
		mov edi, ehci_rh_destroy ; function pointer
		mov [esi+USBDEV.destroy], edi
		mov edi, ehci_rh_poll ; function pointer
		mov [esi+USBDEV.poll], edi
		mov edi, [esi+USBDEV.controller]
		mov edi, [edi+EHCI.op]
endp

; Destroy Root Hub
; ----------------
; ehci_rh_destroy(usbdev_t *dev)
proc ehci_rh_destroy stdcall, dev:dword
endp

; Poll Root Hub
; -------------
; ehci_rh_poll(usbdev_t *dev)
proc ehci_rh_poll stdcall, dev:dword
endp

; Hand over port for companion controller
; ---------------------------------------
; ehci_rh_hand_over_port(usbdev_t *dev, u32 port)
proc ehci_rh_hand_over_port stdcall, dev:dword, port:dword
        mov esi, [dev]
		mov esi, [esi+USBDEV.data]
		mov esi, [esi+EHCI_RH.ports]
		lea esi, [esi+eax*4] ; calculate port offset
		mov eax, [port]
; Clear ConnectStatusChange before execution
		mov edi, [esi] ; read port register value
		or  edi, EHCI_P_CONN_STATUS_CHANGE
		mov [esi], edi ; write it back
; Lowspeed device. Hand over to companion controller
		mov edi, [esi] ; read port register value
		or  edi, EHCI_P_PORT_OWNER
		mov [esi], edi ; write it back
; Wait 100ms and testing for
; status change all the time
        mov ecx, 10
.loop:
		mov edi, [esi] ; read port register value
		test edi, EHCI_P_CONN_STATUS_CHANGE
		jz .changed
		; timeout 10ms
		loop .loop
.changed:
		cmp ecx, 0
		jz  .err ; timeout
		mov edi, [esi] ; read port register value
		; RW/C register, clear it by writing 1
		or  edi, EHCI_P_CONN_STATUS_CHANGE
		mov [esi], edi ; write it back
		ret
.err:
		mov eax, 1
		ret
endp

; Scan port for status changes
; ----------------------------
; ehci_rh_scanport(usbdev_t *dev, u32 port)
proc ehci_rh_scanport, dev:dword, port:dword
		mov esi, [dev]
		mov esi, [esi+USBDEV.data]
		mov eax, [port]
		mov edi, [esi+EHCI_RH.devices+eax*4]
		cmp edi, -1
		jz .handle
.detach:
		mov esi, [dev]
		mov edi, [esi+USBDEV.data]
		stdcall usb_detach_device, [esi+USBDEV.controller], [edi+EHCI_RH.devices+eax*4]
		mov [edi+EHCI_RH.devices+eax*4], -1
.handle:
; Check if PORT_OWNER bit set and clear it,
; to reset all stuff before
		mov esi, [dev]
		mov esi, [esi+USBDEV.data]
		lea esi, [esi+EHCI_RH.ports+eax*4]
		mov edi, [esi] ; read port status value
		test edi, EHCI_P_PORT_OWNER
		jz .cont
		or edi, not EHCI_P_PORT_OWNER
		mov [esi], edi ; write it back
.cont:
		mov edi, [esi] ; read port status again - do not remove!
		test edi, EHCI_P_CURR_CONN_STATUS
		jz .nothing
		; delay 100ms - USB2.0 specification, 9.1.2 chapter
		mov edi, [esi] ; read it again
		and edi, EHCI_P_LINE_STATUS
		cmp edi, EHCI_P_LINE_STATUS_LOWSPEED
		jnz .cont2
		sdtcall ehci_rh_hand_over_port, [dev], [port]
		cmp eax, 0
		jnz .return
.cont2:
; Deassert enable, assert reset.
; These must change automatically
		mov edi, [esi] ; read port status again
		and edi, not EHCI_P_PORT_ENABLE
		or edi, EHCI_P_PORT_RESET
		mov [esi], edi ; and write it back
		; Wait a bit while reset is active
		; delay 50ms - USB2.0 specification 7.1.7.5 chapter
; Deassert reset
		mov edi, [esi] ; read port status
		and edi, not EHCI_P_PORT_RESET
		mov [esi], edi ; and write it back
; Wait max 2ms (EHCI spec 2.3.9)
; for flag change to finish
		mov ecx, 20
.loop2:
		mov edi, [esi] ; read port status
		test edi, EHCI_P_PORT_RESET
		jnz .cont3
		;delay 100us
		loop .loop2
; Check if port reset successful
		mov edi, [esi] ; read port status
		test edi, EHCI_P_PORT_RESET
		jz .err
		; TODO: add check for full/high speed
		mov esi, [dev]
		stdcall usb_attach_device, [esi+USBDEV.controller], [esi+USBDEV.addr], [port], 2
		mov esi, [esi+USBDEV.data]
		mov ebx, [port]
		mov esi, [esi+EHCI_RH.devices+ebx*4]
		mov [esi], eax
.nothing:
		mov esi, [dev]
		mov eax, [port]
		mov esi, [esi+USBDEV.data]
; RW/C register, so clear it by writing 1
		lea esi, [esi+EHCI_RH.ports+eax*4]
		mov edi, [esi] ; read port status register
		or  edi, EHCI_P_CONN_STATUS_CHANGE
		mov [esi], edi
.return:
		ret
.err:
		; timeout error
		ret
endp

; ---------------------------------------------------------------
;
;                         Transactions

; +=====================================================+
; |###################### EHCI QH ######################|
; +-----------------------------------------------------+
; |  Horizontal Link Pointer                            |
; +-----------------------------------------------------+
; |           |          Maximum Packet Length          |
; |     E     |            Device Address               |
; |     P     |         Inactivate on Next Transaction  |
; |     C     |            Endpoint Number              |
; |     H     |               Endpoint Speed            |
; |     A     |           Data Toggle Control           |
; |     R     |    Head of Reclamation List Flag        |
; |           |        Control Endpoint Flag            |
; |           |           Nak Count Reload              |
; +-----------------------------------------------------+
; |  Current QTD pointer                                |
; +-----------------------------------------------------+
; |:::::::::::::::::::: QTD OVERLAY ::::::::::::::::::::|
; |..[IN/OUT/SETUP].....................................|
; |..
;                      .........
; +-----------------------------------------------------+
; |:::::::::::::::::::: QH QTD CHAIN :::::::::::::::::::|
;                         .....
; +=====================================================+
; |::::::::::::::::::: QTD OVERLAY :::::::::::::::::::|
; +---------------------------------------------------+
; |..[OUT]............................................|
; |:|============ EHCI TD at [0x0004c7f0] ==========|:|
; |:| ERRORS = [3] | TOKEN = [0x00000000] |         |:|
; |:+-----------------------------------------------+:|
; |:| Next qTD        [0x0004c6c0]                  |:|
; |:+-----------------------------------------------+:|
; |:| Alt. Next qTD   [0x00000001]                  |:|
; |:+-----------------------------------------------+:|
; |:|       | Bytes to Transfer            | [0000] |:|
; |:|       | PID CODE:                    |    [0] |:|
; |:|       | Interrupt On Complete (IOC)  |    [0] |:|
; |:|       | Status Active                |    [0] |:|
; |:|       | Status Halted                |    [0] |:|
; |:| TOKEN | Status Data Buffer Error     |    [0] |:|
; |:|       | Status Babble detected       |    [0] |:|
; |:|       | Status Transaction Error     |    [0] |:|
; |:|       | Status Missed Micro Frame    |    [0] |:|
; |:|       | Split Transaction State      |    [0] |:|
; |:|       | Ping State                   |    [0] |:|
; |:|-----------------------------------------------|:|
; |...................................................|

; EHCI Bulk Transaction
; ---------------------
; int ehci_bulk(endpoint_t *ep, int size, void *data, int finalize)
proc ehci_bulk stdcall, ep:dword, size:dword, dat:dword, finalize:byte
		locals
			chunk dd 0
			endp dd 0
			hubaddr dd 0
			hubport dd 0
			mlen dd 0
			qh dd 0
			qtd_head dd 0
			qtd_current dd 0
			pid dd 0
			result dd 0
			size1 dd 0
			dat1 dd 0
		endl

		mov esi, [ep]
		mov eax, [esi+ENDPOINT.endpoint]
		and eax, 0xf
		mov [endp], eax
		mov eax, [esi+ENDPOINT.maxpacketsize]
		mov [mlen], eax
		mov eax, [esi+ENDPOINT.direction]
		cmp eax, DIR_IN
		jnz .dir_out
		mov [pid], EHCI_IN

	.dir_out:
		mov [pid], EHCI_OUT
		mov edi, [esi+ENDPOINT.device]
		mov eax, [edi+USBDEV.speed]
		cmp eax, 2
		jge .continue
		;start split transaction
		stdcall ehci_closest_usb2_hub, [dev], [hubaddr], [hubport]
		cmp eax, 0
		jnz .ret1

	.continue:
		stdcall memalign, 32, EHCI_QTD.sizeof
		mov [qtd_head], eax
		mov [qtd_current], eax
		mov ebx, [size]
		mov [size1], ebx
		mov ebx, [dat]
		mov [dat1], ebx

	.fill_loop:
		mov esi, [qtd_current]
		stdcall memset, [qtd_current], 0, EHCI_QTD.sizeof
		mov eax, [pid]
		shl eax, EHCI_QTD_PID_SHIFT
		or eax, (EHCI_QTD_ACTIVE or (0 shl EHCI_QTD_CERR_SHIFT))
		mov esi, [qtd_current]
		mov [esi+EHCI_QTD.token], eax
		stdcall echi_fill_qtd, [qtd_current], [dat], [size]
		mov [chunk], eax
		mov ebx, [size1]
		sub ebx, [chunk]
		mov [size1], ebx
		mov ebx, [dat1]
		add ebx, [chunk]
		mov [dat1], ebx
		mov [esi+EHCI_QTD.alt_next_qtd], EHCI_QTD_TERMINATE
		cmp [size1], 0
		jnz .fill_next
		stdcall get_pg_address, 0
		or eax, EHCI_QTD_TERMINATE
		mov [esi+EHCI_QTD.next_qtd], eax
		jmp .create_qh

	.fill_next:
		stdcall memalign, 32, EHCI_QTD.sizeof
		mov [qtd_next], eax
		stdcall get_pg_addr, [qtd_next]
		mov [esi+EHCI_QTD.next_qtd], eax
		mov esi, [qtd_next]
		mov [qtd_current], esi
		loop .fill_loop

	.create_qh:
		stdcall memalign, 32, EHCI_QH.sizeof
		mov [qh], eax
		stdcall memset, [qh], 0, EHCI_QH.sizeof
		mov esi, [qh]
		stdcall get_pg_addr, [qh]
		or eax, EHCI_QH_QH
		mov [esi+EHCI_QH.horiz_link_ptr], eax

		; filling EPCHAR field
		mov edi, [ep]
		mov edi, [edi+ENDPOINT.device]
		mov eax, [edi+USBDEV.address]
		mov ebx, [endp]
		shl ebx, EHCI_QH_EP_SHIFT
		or eax, ebx
		mov ebx, [edi+USBDEV.speed]
		shl ebx, EHCI_QH_EPS_SHIFT
		or eax, ebx
		or eax, ((0 shl EHCI_QH_DTC_SHIFT) or (1 shl EHCI_QH_RECLAIM_HEAD_SHIFT) or (0 shl EHCI_QH_NAK_CNT_SHIFT))
		mov ebx, [mlen]
		shl ebx, EHCI_QH_MPS_SHIFT
		or eax, ebx
		mov [esi+EHCI_QH.epchar], eax

		; filling EPCAPS field
		mov eax, (3 shl EHCI_QH_PIPE_MULTIPLIER_SHIFT)
		mov ebx, [hubaddr]
		shl ebx, EHCI_QH_HUB_ADDRESS_SHIFT
		or eax, ebx
		mov ebx, [hubport]
		shl ebx, EHCI_QH_PORT_NUMBER_SHIFT
		or eax, ebx
		mov [esi+EHCI_QH.epcaps], eax

		; filling up QH qTD overlay
		stdcall get_pg_addr, [qtd_head]
		mov [esi+EHCI_QH.td+EHCI_QTD.next_qtd], eax
		mov [esi+EHCI_QH.td+EHCI_QTD.alt_next_qtd], ECHI_QTD_TERMINATE
		mov edi, [ep]
		mov eax, [edi+ENDPOINT.toggle]
		cmp eax, 0
		jg .toggle_data1
		or [esi+EHCI_QH.td+EHCI_QTD.token], 0
		mov esi, [qtd_head]
		or [esi+EHCI_QTD.token], 0

	.toggle_data1:
		or [esi+EHCI_QH.td+EHCI_QTD.token], EHCI_QTD_TOGGLE_DATA1
		mov esi, [qtd_head]
		or [esi+EHCI_QTD.token], EHCI_QTD_TOGGLE_DATA1

		mov edi, [edi+ENDPOINT.device]
		stdcall ehci_process_async_schedule, [edi+EHCI.controller], [qh], [qtd_head]
		mov [result], eax
		stdcall ehci_free_qh_and_qtds, [qh], [qtd_head]
		mov eax, [result]
		ret

	.ret1: ; return 1
		mov eax, 1
		ret
endp

; EHCI Control Transaction
; ------------------------
; int ehci_control(usbdev_t *dev, int direction, int drlen,
;       void *devreq, int dalen, void *data)
proc ehci_control stdcall, dev:dword, dir:byte, drlen:dword,
        devreq:dword, dalen:dword, dat:dword
		locals
			endp dd 0
			mlen dd 0
			hubaddr dd 0
			hubport dd 0
			non_hs_ctrl_ep dd 0
			qtd_head dd 0
			qtd_current dd 0
			qtd_next dd 0
			qh dd 0
			result dd 0
		endl

		mov esi, [dev]
		mov eax, [esi+USBDEV.endpoints + ENDPOINT.maxpacketsize] ; dev->endpoints[0].maxpacketsize
		mov [mlen], eax
		mov eax, [esi+USBDEV.speed]
		cmp eax, 2 ; test, if this high-speed USB2.0 device
		jnz .continue
		stdcall ehci_closest_usb2_hub, [dev], [hubaddr], [hubport]
		cmp eax, 0
		jnz .ret1
		mov [non_hs_ctrl_ep], 1

	.create_qtds:
		stdcall memalign, 32, EHCI_QTD.sizeof
		mov [qtd_head], eax
		mov [qtd_current], eax
		stdcall memset, [qtd_current], 0, EHCI_QTD.sizeof
		mov esi, [qtd_current]
		mov [esi+EHCI_QTD.token], (EHCI_QTD_ACTIVE or (EHCI_SETUP shl EHCI_QTD_PID_SHIFT) or (3 shl EHCI_QTD_CERR_SHIFT)
		; data toggle is disabled so do not mask it
		stdcall ehci_fill_qtd, [qtd_current], [devreq], [drlen]

if ehci_debug > 0
		cmp eax, [drlen]
		jz .create_next
		; debug message - couldn't send full device request
	.create_next:
end if

		; create next qTD in QH
		stdcall memalign, 32, EHCI_QTD.sizeof
		mov [qtd_next], eax
		; calculate physical address from virtual
		stdcall get_pg_addr, [qtd_next]
		mov [esi+EHCI_QTD.next_qtd], eax
		mov [esi+EHCI_QTD.alt_next_qtd], EHCI_QTD_TERMINATE
		; fill qTD chain here
		cmp [dalen], 1
		jle .create_last
		mov eax, [qtd_next]
		mov [qtd_current], eax
		stdcall memset, [qtd_current], 0, EHCI_QTD.sizeof
		mov esi, [qtd_current]
		cmp [dir], DIRECTION_OUT
		jz  .dir_out
		mov eax, (EHCI_IN shl EHCI_QTD_PID_SHIFT)

	.dir_out:
		mov eax, (EHCI_OUT shl EHCI_QTD_PID_SHIFT)
		mov [esi+EHCI_QTD.token], (EHCI_QTD_ACTIVE or EHCI_QTD_TOGGLE_DATA1 or (3 shl EHCI_QTD_CERR_SHIFT) or eax)
		stdcall ehci_fill_qtd, [qtd_current], [dat], [dalen]

if ehci_debug > 0
		cmp eax, [dalen]
		jz .close_next
		; debug message - couldn't send entire control payload
	.close_next:
end if

		stdcall memalign, 32, EHCI_QTD.sizeof
		mov [qtd_next], eax
		mov esi, [qtd_current]
		; calculate physical address from virtual
		stdcall get_pg_addr, [qtd_next]
		mov [esi+EHCI_QTD.next_qtd], eax
		mov [esi+EHCI_QTD.alt_next_qtd], EHCI_QTD_TERMINATE

	.create_last:
		; create last qTD in QH
		mov eax, [qtd_next]
		mov [qtd_current], eax
		stdcall memset, [qtd_current], 0, EHCI_QTD.sizeof
		mov esi, [qtd_current]
		cmp [dir], DIRECTION_OUT
		jz .dir_out2
		mov eax, (EHCI_OUT shl EHCI_QTD_PID_SHIFT)

	.dir_out2:
		mov eax, (EHCI_IN shl EHCI_QTD_PID_SHIFT)
		mov [esi+EHCI_QTD.token], (EHC_QTD_ACTIVE or EHCI_QTD_TOGGLE_DATA1 or (0 shl EHCI_QTD_CERR_SHIFT) or eax)
		stdcall ehci_fill_qtd, [qtd_current], 0, 0
		mov [esi+EHCI_QTD.next_qtd], ECHI_QTD_TERMINATE
		mov [esi+EHCI_QTD.alt_next_qtd], EHCI_QTD_TERMINATE
		; create QH
		stdcall memalign, 32, EHCI_QH.sizeof
		mov [qh], eax
		stdcall memset, [qh], 0, EHCI_QH.sizeof
		mov esi, [qh]

		; calculate physical address from virtual
		stdcall get_pg_addr, [qh]
		or eax, EHCI_QH_QH
		mov [esi+EHCI_QH.horiz_link_ptr], eax

		; filling EPCHAR field
		mov edi, [dev]
		mov eax, [edi+USBDEV.address]
		mov ebx, [endp]
		shl ebx, EHCI_QH_EP_SHIFT
		or eax, ebx
		mov ebx, [edi+USBDEV.speed]
		shl ebx, EHCI_QH_EPS_SHIFT
		or eax, ebx
		or eax, ((1 shl EHCI_QH_DTC_SHIFT) or (1 shl EHCI_QH_RECLAIM_HEAD_SHIFT))
		mov ebx, [mlen]
		shl ebx, EHCI_QH_MPS_SHIFT
		or eax, ebx
		mov ebx, [non_hs_ctrl_ep]
		shl ebx, EHCI_QH_NON_HS_CTRL_EP_SHIFT
		or eax, ebx
		mov [esi+EHCI_QH.epchar], eax

		; filling EPCAPS field
		mov eax, (3 shl EHCI_QH_PIPE_MULTIPLIER_SHIFT)
		mov ebx, [hubaddr]
		shl ebx, EHCI_QH_HUB_ADDRESS_SHIFT
		or eax, ebx
		mov ebx, [hubport]
		shl ebx, EHCI_QH_PORT_NUMBER_SHIFT
		or eax, ebx
		mov [esi+EHCI_QH.epcaps], eax

		stdcall get_pg_addr, [qtd_head]
		; fill the QH overlay
		mov [esi+EHCI_QH.td+EHCI_QTD.next_qtd], eax
		mov [esi+EHCI_QH.td+EHCI_QTD.alt_next_qtd], EHCI_QTD_TERMINATE
		stdcall ehci_process_async_schedule, [edi+EHCI.controller], [qh], [qtd_head]
		mov [result], eax

		; free all memory for this transaction
		stdcall ehci_free_qh_and_qtds, [qh], [qtd_head]
		mov eax, [result]
		ret

	.ret1: ; return 1;
		mov eax, 1
		ret
endp

; vim: ft=fasm

